/* =============================================================
   PACHET DE FUNCTII PENTRU INTERPOLAREA POLINOMIALA PE PORTIUNI

   Data: 06.08.1991
   Data ultimei modificari: 08.08.1991
   Programatori: Daniel Ioan, Irina Munteanu
   Functii:
	- tridiag - rezolva sisteme liniare cu matrice tridiagonala
	- cauta_seg  - determina segmentul in care se afla valoarea variabilei
	  curente
	- eval_l - evalueaza functia de interpolare liniara pe portiuni
	- prep_h - pregateste datele pentru interpolarea Hermite
	- prep_bess - pregateste datele pentru interpolarea Bessel
	- prep_akima - pregateste datele pentru interpolarea Akima
	- prep_spln - pregateste datele pentru spline cubic natural
	- prep_spla - pregateste datele pentru spline cubic aproximativ
	- eval_h - evalueaza functia spline cubica de interpolare Hermite
	- eval_b - evalueaza functia de interpolare cu metoda B-spline cubica
	- asezare - reaseaza nodurile pentru minimizarea erorii
   =============================================================== */
#define fmax(a,b) ((a>b)?a:b)

/* Rezolvarea sistemelor tridiagonale */
/*------------------------------------*/
/* Data : 28.01.90
 Rezolva un sistem de ecuatii cu matricea A tridiagonala.
 Date de intrare:
		  n = numar de eecuatii
		  p[N] = vectorul subdiagonal
		  q[N] = vectorul diagonalei;  Nu se accepta q[i] nul.
		  r[N] = vectorul de deasupra diagonalei.
		  b[N] = termenul liber
 Date de iesire:
		  x[N] = solutia.
*/

tridiag(int n, float p[N], float q[N], float r[N], float b[N], float x[N])
{

  int k;
  float s;

  /* Eliminare */
  for (k=1; k<n; k++)
    {
    if (q[k-1] == 0) return 1;	/* eroare */
    s = p[k]/q[k-1];
    q[k] -= r[k-1]*s;
    b[k] -= b[k-1]* s;
    }
  /* Substitutie */
  if (q[n-1] == 0) return 1; 	/* eroare */
  x[n-1] = b[n-1]/q[n-1];
  for (k=n-2; k>=0; k--)
    x[k] = (b[k] - r[k]*x[k+1]) / q[k];

  return 0;	/* all OK */
}


/*===============================================================
  Determina segmentul in care se afla valoarea variabilei curente
  prin metoda cautarii binare
  =============================================================== */

int cauta_seg(int n,	/* numar de segmente */
			  float x[N],	/* abscisele nodurilor */
			  float xcrt)	/* valoarea variabilei curente */
{
int k,	/* valoarea intoarsa - segmentul numerotat dupa nodul initial */
	k1,k2;	/* limitele intervalului curent */

if (xcrt >= x[n-1]) return (n-1);
if (xcrt <= x[1]) return (0);

k1 = 1;
k2 = n-1;
do
	{
	k = (k1+k2)/2;
	if (xcrt < x[k]) k2 = k;
	else k1 = k;
	} while ((k2 - k1) > 1);

return (k1);
}	/* End cauta_seg */

/*====================================================
  Evalueaza functia de interpolare liniara pe portiuni
  ==================================================== */
float eval_l(int n,		/* numar de intervale */
	   float x[N],
	   float y[N],	/* datele de interpolat */
	   float xcrt)	/* valoarea variabilei independente */
{

int k;	/* interval curent */

k = cauta_seg(n,x,xcrt);
return (y[k] + (y[k+1] - y[k])/(x[k+1] - x[k]) * (xcrt - x[k]));

}	/* End eval_l */

/*=============================================
  Pregateste datele pentru interpolarea Hermite
  Calculeaza coeficientii polinoamelor cubice pe fiecare interval,
  folosind valorile functiei si ale derivatei la capetele intervalului
  ============================================= */
void prep_h(int n,		/* numar intervale */
	   float x[N],	/* abscisele nodurilor */
	   float y[N],	/* valorile functiei */
	   float yd[N],	/* derivatele in noduri */
	   float c[4][N])	/* coeficientii polinoamelor */
{
int k;
float dx,dy;

for (k=0; k<n; k++)
	{
	c[0][k] = y[k];
	c[1][k] = yd[k];
	dx = x[k+1] - x[k];
	dy = (y[k+1] - y[k])/dx;
	c[3][k] = (yd[k+1] + yd[k] - 2*dy)/dx;
	c[2][k] = (dy - yd[k])/dx - c[3][k];
	c[3][k] /= dx;
	}
}	/* End prep_h */

/*=============================================
  Pregateste datele pentru interpolarea Bessel
  Calculeaza numeric derivatele functiei in nodurile retelei
  folosind interpolare locala parabolica
  ============================================= */
void prep_bess(int n,	/* numar intervale */
	   float x[N],	/* abscisele nodurilor */
	   float y[N],	/* valorile functiei */
	   float yd[N])	/* derivatele in noduri - rezultate */
{
int k;
float dy[N];	/* tabelul diferntelor divizate */

for (k=0; k<n; k++)
	dy[k] = (y[k+1] - y[k])/ (x[k+1] - x[k]);
yd[0] = dy[0] + (dy[1] - dy[0])/(x[2] - x[0])*(x[0] - x[1]);
yd[n] = dy[n-1] + (dy[n-1] - dy[n-2])/(x[n] - x[n-2])*(x[n] - x[n-1]);
for (k=1; k<n; k++)
	yd[k] = (dy[k-1]* (x[k+1] - x[k])
	+ dy[k]*(x[k] - x[k-1]))/ (x[k+1] - x[k-1]);

}	/* End prep_bess */

/*=============================================
  Pregateste datele pentru interpolarea Akima
  Calculeaza numeric derivatele functiei in nodurile retelei
  folosind interpolare locala parabolica
  ============================================= */
void prep_akima(int n,	/* numar intervale */
	   float x[N],	/* abscisele nodurilor */
	   float y[N],	/* valorile functiei */
	   float yd[N])	/* derivatele in noduri - rezultate */
{
int k;
float dy[N];	/* tabelul diferntelor divizate */
float wp,wm;

for (k=0; k<n; k++)
	dy[k] = (y[k+1] - y[k])/ (x[k+1] - x[k]);
yd[0] = dy[0] + (dy[1] - dy[0])/(x[2] - x[0])*(x[0] - x[1]);
yd[n] = dy[n-1] + (dy[n-1] - dy[n-2])/(x[n] - x[n-2])*(x[n] - x[n-1]);
dy[n] = dy[n-1];
yd[1] = (dy[0]* (x[2] - x[1]) + dy[1]*(x[1] - x[0]))/ (x[2] - x[0]);
yd[n-1] = (dy[n-2]* (x[n] - x[n-1]) +
		  dy[n-1]*(x[n-1] - x[n-2]))/ (x[n] - x[n-2]);
	/* Pentru extremitati s-a folosit formula in 3 puncte Bessel */
for (k=2; k<n-1; k++)
	{
	wp = fabs(dy[k+1] - dy[k]);
	wm = fabs(dy[k-2] - dy[k-1]);
	if (wm + wp != 0)
		yd[k] = (dy[k-1]* wp + dy[k]*wm)/ (wp+wm);
	else
		yd[k] = (dy[k-1] + dy[k])/2;
	}

}	/* End prep_akima */


/*============================================================
  Pregateste datele pentru interpolarea spline cubica naturala
  Calculeaza numeric derivatele functiei in nodurile retelei
  prin rezolvarea unui sistem tridiagonal obtinut din
  conditiile de continuitate a derivatei a doua. La capetele
  domeniului de definitie derivata a doua se considera nula.
  ============================================================ */
void prep_spln(int n,	/* numar intervale */
	   float x[N],	/* abscisele nodurilor */
	   float y[N],	/* valorile functiei */
	   float yd[N])	/* derivatele in noduri - rezultate */
{
int k;
float p[N],q[N],r[N],	/* diagonalele matricei */
	  s[N];				/* termenul liber */

for (k=0; k<n; k++)
	{
	p[k] = x[k+1] - x[k];
	q[k] = (y[k+1] - y[k])/ p[k];
	}
for (k=1; k<n; k++)
	s[k] = 3.* (q[k-1]*p[k] + q[k]*p[k-1]);
s[0] = 3.*q[0];
s[n] = 3.*q[n-1];
for (k=1; k<n; k++)
	{
	q[k] = 2.*(x[k+1] - x[k-1]);
	r[k] = p[k-1];
	}
q[0] = 2.;
r[0] = 1.;
p[n] = 1.;
q[n] = 2.;
tridiag(n+1,p,q,r,s,yd);

}	/* End prep_spln */


/*============================================================
  Pregateste datele pentru interpolarea spline cubica aproximativa.
  Calculeaza numeric derivatele functiei in nodurile retelei
  prin rezolvarea unui sistem tridiagonal obtinut din
  conditiile de continuitate a derivatei a doua. La capetele
  domeniului de definitie derivata se evalueaza numeric.
  ============================================================ */
void prep_spla(int n,	/* numar intervale */
	   float x[N],	/* abscisele nodurilor */
	   float y[N],	/* valorile functiei */
	   float yd[N])	/* derivatele in noduri - rezultate */
{
int k;
float p[N],q[N],r[N],	/* diagonalele matricei */
	  s[N];				/* termenul liber */

for (k=0; k<n; k++)
	{
	p[k] = x[k+1] - x[k];
	q[k] = (y[k+1] - y[k])/ p[k];
	}
for (k=1; k<n; k++)
	s[k] = 3.* (q[k-1]*p[k] + q[k]*p[k-1]);
s[0] = 3.*q[0];
s[n] = 3.*q[n-1];
for (k=1; k<n; k++)
	{
	q[k] = 2.*(x[k+1] - x[k-1]);
	r[k] = p[k-1];
	}
q[0] = 3.;
r[0] = 0.;
p[n] = 0.;
q[n] = 3.;
tridiag(n+1,p,q,r,s,yd);

}	/* End prep_spla */


/*=====================================================
  Evalueaza functia spline cubica de interpolare Hermite
  ====================================================== */
float eval_h(int n,		/* numar intervale */
	   float x[N],		/* abscisele nodurilor */
	   float c[4][N],	/* coeficientii polinoamelor */
	   float xcrt)		/* valoarea variabilei independente */
{
int k;          /* indice interval in care se afla xcrt */
float ycrt;		/* valoarea cautata a functiei */
int i;

k = cauta_seg(n, x, xcrt);
ycrt = c[3][k];
for (i=2; i>=0; i--)
	ycrt = c[i][k] + ycrt* (xcrt - x[k]);
return (ycrt);

}	/* End eval_h */


/*==========================================================
  Evalueaza functia de interpolare cu metoda B-spline cubica
  ========================================================== */
float eval_b(int n,		/* numar intervale */
	   float x[N],		/* abscisele nodurilor */
	   float y[N],		/* valorile functiei */
	   float yd[N],		/* derivatele in noduri */
	   float xcrt)		/* valoarea variabilei independente */
{
int k;          /* indice interval in care se afla xcrt */
float ycrt;		/* valoarea cautata a functiei */
float u,v,u2,v2;

k = cauta_seg(n, x, xcrt);
u = (xcrt - x[k]) / (x[k+1]-x[k]);
v = 1 - u;
u2 = u*u;
v2 = v*v;
ycrt = v2 * (yd[k]* (xcrt - x[k]) + y[k]*(3.-2*v)) +
	   u2*(yd[k+1]* (xcrt - x[k+1]) + y[k+1]* (3.-2*u));
return (ycrt);

}	/* End eval_b */

/*==========================================================
  Reaseaza nodurile pentru minimizarea erorii la interpolarea
  liniara pe portiuni
  ========================================================== */
float asezare(int n,		/* numar intervale */
	   float x[N],		/* abscisele nodurilor */
	   float y[N])		/* valorile functiei */
{
int k;
float gg[N];
float e[N];
float err,eps;

/* Evalueaza prima derivata */
for (k=0; k<n; k++)
	e[k] = (y[k+1] - y[k])/(x[k+1]-x[k]);
/* Evalueaza radicalul derivatei a doua */
for (k=1; k<n; k++)
	gg[k] = sqrt(fabs((e[k]-e[k-1])/(x[k+1] - x[k-1])));
gg[0] = gg[1];
gg[n] = gg[n-1];
e[0]  = 0.;

/* Evalueaza integrala din radicalul derivatei secunde */
for (k=1; k<=n; k++)
	e[k] = e[k-1] + (x[k]-x[k-1])*(gg[k]+gg[k-1])/2.;

/* Evalueaza eroarea maxima la interpolarea liniara */
	eps = 0;
	for (k=0; k<n; k++)
		{
		err = fmax(gg[k], gg[k+1]) * (x[k+1] - x[k]);
		if (err>eps) eps = err;
		}
	eps = eps*eps/8.;
/*	printf ("\n%f \n", eps);
 */
/* Determina pozitia optima a nodurilor prin interpolare inversas */
for (k=1; k<n; k++)
	{
	gg[k] = eval_l(n,e,x,(float)k*e[n]/n);
	y[k] = eval_l(n,x,y,gg[k]); /* reactualizeaza y */
	}

/* Actualizeaza noduri  */
for (k=1; k<n; k++)
	x[k] = gg[k];

return (eps);

}	/* End asezare */

