/****************************************************************************
 *                                                                          *
 *                                                                          *
 *     Pachetul de functii permite:                                         *
 *       - calcularea coeficientilor Fourier ai unei functii                *
 *         aproximata prin trapeze, periodica de perioada T;                *
 *       - calculul sumei partiale a m termeni din descompunerea            *
 *         Fourier a functiei aproximata prin trapeze;                      *
 *                                                                          *
 *                                                                          *
 *     Autor: Mihai Lazarescu                                               *
 *     Data : 17.11.1991                                                    *
 *                                                                          *
 *     (C) 1991 L.M.N., Mihai Lazarescu                                     *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/


/* Fisiere incluse */
#include <math.h>
#include "fourier.h"


/* FUNCTII */
/* Descompune in serie Fourier o functie periodica liniara pe portiuni */
void Fourier_lpp(int n, double x[], double y[], int m, double a[], double b[], double w)

/*
   Functia determina primii m coeficienti Fourier ak in sin si bk in cos ai
   unei functii periodice liniare pe portiuni descrisa prin cele
   n puncte de fringere cu coordonatele xi, yi.
   x0 trebuie sa fie nul, iar xn-1 egal cu perioada functiei.
   Pentru descrierea discontinuitatilor se admit si noduri confundate.
   Aproximarea Fourier a functiei este:
                 m
     f(t) = b0 +  [ak*sin(k*w*t) + bk*cos(k*w*t)]
                i=1
*/

{
    double       s1, s2;
    double       c1, c2;
    double       p;
    const float  PI  = 3.14159265;
    const double EPS = 1e-15;
    int          i, k;

    /* Initializari */
    a[0] = b[0] = 0.0;

    /* Calculeaza valoarea medie a functiei */
    for (i = 0; i < n-1; i++)
        b[0] += (y[i+1]+y[i]) * (x[i+1]-x[i]);
    b[0] /= 4*PI/w;

    /* Calculul armonicilor superioare */
    for (k = 1; k <= m; k++) {
        double   kw;

        a[k] = b[k] = 0.0;
        kw   = k*w;
        s2   = 0.0;
        c2   = 1.0;
        for (i = 0; i < n-1; i++) {                     /* Parcurge segmentele */
            if (fabs(x[i+1] - x[i]) < EPS) continue;     /* Discontinuitate */
            s1    = s2;
            c1    = c2;
            s2    = sin(kw*x[i+1]);
            c2    = cos(kw*x[i+1]);
            p     = (y[i+1]-y[i]) / (x[i+1]-x[i]);
            a[k] += (y[i]*c1   - y[i+1]*c2) + p*(s2 - s1)/kw;
            b[k] += (y[i+1]*s2 - y[i]*s1)   + p*(c2 - c1)/kw;
        }
        a[k] /= k*PI;
        b[k] /= k*PI;
    }
}

/* Sumarea unei serii Fourier trunchiate */
void eval_Fourier(int n, double x[], double y[], int m, double a[], double b[], double w)

/*   Calculeaza seria Fourier trunchiata cu m armonici */
/* care are coeficientii ak in sin si bk in cos. */
/*   Evaluarea este efectuata in n+1 puncte date */
/* x0, x1, ..., xn, iar valorile intoarse in tabloul y. */

{
    int     i, k;       /* Contoare */
    double  kw = w;

    /* Initializari */
    for (i = 0; i < n; i++)
        y[i] = b[0];

    for (k = 1; k <= m; k++) {
        for (i = 0; i < n; i++)
            y[i] += a[k]*sin(kw*x[i]) + b[k]*cos(kw*x[i]);
        kw += w;
    }
}
