/****************************************************************************
 *                                                                          *
 *                                                                          *
 *     Pachetul de functii permite:                                         *
 *       - calcularea coeficientilor Fourier ai unei functii                *
 *         aproximata prin trapeze, periodica de perioada T;                *
 *       - calculul sumei partiale a m termeni din descompunerea            *
 *         Fourier a functiei aproximata prin trapeze;                      *
 *       - calculul transformatei Fourier discrete directe;                 *
 *       - calculul transformatei Fourier discrete inverse;                 *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *     Autor: Mihai Lazarescu                                               *
 *     Data : 21.11.1991                                                    *
 *                                                                          *
 *     (C) 1991 L.M.N., Mihai Lazarescu                                     *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/


// Fisiere incluse
#include <complex.h>
#include "fourierx.h"



// FUNCTII
// Descompune in serie Fourier o functie periodica liniara pe portiuni
void Fourier_lpp(int n, float t[], complex f[], int m, complex c[])

// Functia determina primii m coeficienti Fourier ck ai
// unei functii periodice liniare pe portiuni descrisa prin cele
// n puncte de fringere cu coordonatele ti, fi.
// t0 trebuie sa fie nul, iar tn egal cu perioada functiei.
// Pentru descrierea discontinuitatilor se admit si noduri confundate.
// Aproximarea Fourier a functiei este:
//                     m
//   f(t) = Real(c0) +  [Imag(ck)*sin(k*w*t) + Real(ck)*cos(k*w*t)]
//					  i=1

{
	float	    w;
	float	    s1, s2;
	float	    c1, c2;
	float	    p;        	        // Variabile auxiliare
    const float PI  = 3.14159265;
    const float EPS = 1e-6;
	int		    i, k;		        // Contoare

	// Initializari
	w    = 2*PI / t[n-1];
	c[0] = complex(0., 0.);

	// Calculeaza valoarea medie a functiei
	for (i = 0; i < n-1; i++)
		c[0] += real(f[i+1]+f[i]) * (t[i+1]-t[i]);
	c[0] /= 2*t[n-1];

	// Calculul armonicilor superioare
	for (k = 1; k <= m; k++) {
		float	kw;

		c[k] = complex(0., 0.);
		kw   = k*w;
		s2   = 0.0;
		c2   = 1.0;
		for (i = 0; i < n-1; i++) {		        		// Parcurge segmentele
			if (fabs(t[i] - t[i+1]) < EPS) continue;	// Discontinuitate
            s1    = s2;
            c1    = c2;
			s2    = sin(kw*t[i+1]);
			c2    = cos(kw*t[i+1]);
			p     = real(f[i+1]-f[i]) / (t[i+1]-t[i]);
			c[k] += complex(real(f[i+1]*s2 - f[i]*s1)   + p*(c2 - c1)/kw,
							real(f[i]*c1   - f[i+1]*c2) + p*(s2 - s1)/kw);
		}
		c[k] /= k*PI;
	}
}

// Sumarea unei serii Fourier trunchiate
void eval_Fourier(int n, float t[], complex f[], int m, complex c[])

// Calculeaza seria Fourier trunchiata cu m armonici
// care are coeficientii Imag(ck) in sin si Real(ck) in cos.
// Evaluarea este efectuata in n puncte date
// t0, t1, ..., tn, iar valorile corespunzatoare sint
// intoarse in fi.

{
	int		i, k;		// Contoare
	float	w, kw;
	float	PI;

	// Initializari
	PI = 4. * atan(1.);
	w = 2*PI / t[n-1];
	for (i = 0; i < n; i++)
		f[i] = real(c[0]);

	for (k = 1; k <= m; k++) {
		kw = k*w;
		for (i = 0; i < n; i++)
			f[i] += imag(c[k])*sin(kw*t[i]) + real(c[k])*cos(kw*t[i]);
	}
}

void transf_F(int dir, int n, complex f[], complex c[])

// Transformata Fourier discreta.
// Daca:
//    - dir=1 - calculeaza transformata directa;
//    - dir=0 - calculeaza transformata inversa.
//  n - numarul datelor de intrare;
//  f - datele de intrare;
//  c - datele de iesire.

{
    long int    i, k;                   // Contoare
    complex     w;
    const float	PI = 3.14159265;

    for (k = 0; k < n; k++)                 // Initializare
        c[(int)k] = 0.0;

    w  = complex(0., 2*PI / n);             // w = j2PI/n
    if (dir) w = -w;

    for (k = 0; k < n; k++) {
        for (i = 0; i < n; i++)
            c[(int)k] += f[(int)i] * exp(w * i*k);
        if (dir) c[(int)k] /= n;
    }
}

void transf_F_opt(int dir, int n, complex f[], complex c[])

// Transformata Fourier discreta (cu optimizarea efortului de calcul).
// Daca:
//   - dir=1 - calculeaza transformata directa;
//   - dir=0 - calculeaza transformata inversa.
// n - numarul datelor de intrare;
// f - datele de intrare;
// c - datele de iesire.

{
    int         i, k;                   // Contoare
    complex	    w;
    complex     e, ek, eki;             // Var. auxiliare
    const float	PI = 3.14159265;

    for (k = 0; k < n; k++)                 // Initializare
        c[k] = 0.0;

    w  = complex(0., 2*PI / n);             // w   =  j2PI/n
    if (dir) w = -w;

    e  = exp(w);                            // e   =  e^w
    ek = 1.0;                               // ek  =  e^(wk)
    for (k = 0; k < n; k++) {
        eki = 1.0;                          // eki = e^(wki)
        for (i = 0; i < n; i++) {
            c[k] += f[i] * eki;
            eki  *= ek;
	    }
        if (dir) c[k] /= n;
        ek *= e;
    }
}

void FFT(complex x[], int n, int inv)

// Implementarea algoritmului FFT pentru calcularea
// coeficientilor Fourier discreti ai unei secvente
// de date de n puncte.

// - n   - numarul de puncte de date;
// - x   - vector complex continind secventa de date la
//         intrare si coeficientii DFT la iesire.
//         Trebuie sa aiba dimensiunea 512.
// - inv - flag pentru inversare:
//         inv=0 pentru transformata directa;
//         inv=1 pentru transformata inversa.

{
#define		PI	3.14159265

	int		iter = 0;
	int		irem = n;
	int		sign = -1;
	int		nxp2 = n;
	int		nxp;
	float	wpwr;
	int		it, m, mxp, i, k, j;	// Contoare
	float	arg;
	complex	w;
	int		j1, j2;
	complex	t;						// Temp.
	int		n1, n2;					// Aux.

//  Calcularea numarului de iteratii (log n in baza 2)

	while ((irem /= 2) != 0)
		iter++;
	if (inv) sign = 1;

//  Calculul la fiecare iteratie:
//	npx  - numarul de puncte intr-o partitie;
//	npx2 - npx/2.

	for (it = 1; it <= iter; it++) {
		nxp   = nxp2;
		nxp2  = nxp/2;
		wpwr  = PI/nxp2;
		for (m = 1; m <= nxp2; m++) {	            // Calculeaza multiplicatorul
			arg = (m-1) * wpwr;
			w   = complex(cos(arg), sign*sin(arg));
			for (mxp = nxp; mxp <= n; mxp += nxp) {	// Calculul pentru fiecare partitie
				j1     = mxp - nxp + m;
				j2     = j1 + nxp2;
				t      = x[j1] - x[j2];
				x[j1] += x[j2];
				x[j2]  = t * w;
			}
		}
	}

//  Reface ordinea valorilor

	n2 = n/2;
	n1 = n - 1;
	j  = 1;
	for (i = 1; i <= n1; i++) {
		if (i < j) {
			t    = x[j];
			x[j] = x[i];
			x[i] = t;
		}
		k = n2;
		while (k < j) {
			j -= k;
			k /= 2;
		}
		j += k;
	}
	if (!inv)
		for (i = 1; i <= n; i++)
			x[i] /= n;
}
